
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "rhp.h"
#include "sif_f.h"
#include "cell.h"
#include "sif.h"

extern void	Error(char *msg);

#ifdef	_WIN32
#include <process.h>
#include <windows.h>

#define THREAD_T	HANDLE
#define FUNCTION_T	unsigned __stdcall
#define RETURN_VALUE	0

#define INIT_THREAD(T,F,A)	\
	(T=(HANDLE)_beginthreadex(NULL,0,F,(void *)(A),0,NULL))==0
#define TERM_THREAD(T)	\
	WaitForSingleObject(T,INFINITE)==WAIT_FAILED || CloseHandle(T)==0
#else
#include <pthread.h>

#define THREAD_T	pthread_t
#define FUNCTION_T	void *
#define RETURN_VALUE	NULL

#define INIT_THREAD(T,F,A)	pthread_create(&(T),NULL,F,(void *)(A))
#define TERM_THREAD(T)		pthread_join(T,NULL)
#endif

#define INIT_MT(T,F,A)	\
if (INIT_THREAD(T,F,A)) Error(#F " : multi-threading initialization error.")
#define TERM_MT(T,F)	\
if (TERM_THREAD(T)) Error(#F " : multi-threading termination error.")

#ifndef M_PI
#define M_PI	3.14159265358979323846
#endif

#ifndef	Float
#define Float	float
#endif

#ifndef	Filter
#define Filter	Ramachandran
#endif

double	Ramachandran(int j)
{
	return (j==0)?1.0/4.0:(j&1)?-1.0/(M_PI*M_PI*(double)j*(double)j):0.0;
}

double	Shepp(int j)
{
	return 2.0/(M_PI*M_PI*(1.0-4.0*(double)j*(double)j));
}

double	Chesler(int j)
{
	if (j==0)
	    return 1.0/8.0-1.0/(2.0*M_PI*M_PI);
	else if (j==1 || j==(-1))
	    return 1.0/16.0-1.0/(2.0*M_PI*M_PI);
	else if (j&1)
	    return -1.0/(2.0*M_PI*M_PI*(double)j*(double)j);
{
	double  j2=(double)j*(double)j,j21=j2-1.0;

	return -(j2+1.0)/(2.0*M_PI*M_PI*j21*j21);
}}

extern double	Filter(int j);

typedef struct {
		double	x,y;
	} Vector;

typedef struct {
		int	j;
		double	F4l,F4Q,u;
	} Struct;

static int	Nu,Nw,N1,N2,N,L,Threads=1;
static double	Ow,l0,**F4P,**P,**Q;
static Vector	*E,*G,**V4T;
static Struct	**S4BP;
static Float	***F;

static void	FFT(int sign,Vector *V)
{
	int	j,k,l,m,p,q;
	Vector	v,*e,*a,*b;
	double	x,y,z;

	for (j=N2-1, k=N2-2; k>0; k--) {
	    for (l=N1; (j^=l)&l; l>>=1) ;

	    if (j>k) {
		v=V[j]; V[j]=V[k]; V[k]=v;
	    }
	}
	for (j=2, k=1, l=N1, m=N2; m>1; m=l, l>>=1, k=j, j<<=1)
	    for (e=E, p=0; p<k; p++, e+=m) {
		x=e->x; y=e->y*(double)sign;
		for (a=V+p, q=0; q<l; q++, a+=j) {
		    v=(*(b=a+k)); b->x=a->x-(z=v.x*x-v.y*y); a->x+=z;
				  b->y=a->y-(z=v.x*y+v.y*x); a->y+=z;
		}
	    }
}

static FUNCTION_T	Convolution(void *i)
{
	size_t	l;
	Vector	*v,*g,w,*V=V4T[(size_t)i];
	double	*f4p,*p;
	int	j;

	for (l=(size_t)i; l<Nw; l+=Threads) {
	    for (v=V, j=0; j<N1; j++, v++) v->x=v->y=0.0;

	    f4p=F4P[l]; p=P[l];
	    for (j=0; j<Nu; j++, v++) {
		v->x=(*(f4p++))*(*(p++)); v->y=0.0;
	    }
	    for (; j<N1; j++, v++) v->x=v->y=0.0;

	    FFT(-1,V);

	    for (v=V, g=G, j=0; j<N2; j++, v++, g++) {
		w=(*v); v->x=g->x*w.x-g->y*w.y;
			v->y=g->x*w.y+g->y*w.x;
	    }
	    FFT(1,V);

	    for (v=V, j=0; j<Nu; j++, v++) Q[j][l]=v->x;
	}
	return RETURN_VALUE;
}

static FUNCTION_T	BackProjection(void *i)
{
	size_t	y;
	int	x,j,z,l;
	Float	**Fy,*Fyx;
	Struct	*s4bp;
	double	*Q0,*Q1,u,F4Q,W,F4l,w,q,Qu,Qw;

	for (y=(size_t)i; y<N; y+=Threads) {
	    Fy=F[y]; s4bp=S4BP[y];
	    for (x=0; x<N; x++)
		if ((j=s4bp[x].j)>=0) {
		    Fyx=Fy[x];
		    Q0=Q[j]; Q1=Q[j+1];
		    u=s4bp[x].u; F4Q=s4bp[x].F4Q;
		    for (W=Ow+(F4l=s4bp[x].F4l)*l0, z=0; z<L; z++, W+=F4l) {
			w=W-(double)(l=(int)W);
			Qu=Q1[l]-(q=Q0[l]); Qw=Q0[++l]-q;
			*(Fyx++)+=(F4Q*(q+Qu*u+Qw*w+(Q1[l]-q-Qu-Qw)*u*w));
		    }
		}
	}
	return RETURN_VALUE;
}

#define LEN	2048
#define EPS	1e-9

int	main(int argc,char **argv)
{
	char		*env,path[LEN];
	HiPic		hp;
	int		M,O,z,j,l,O2,Y,y,O2_Y2,X,x,m,z0=0,bps=0;
	double		A,B,Du,Dw,RC,Ou,u,u1,u2,r,r1,r2,w,Dxy,ra0,s,c,f,v,
			min,max,base,step;
	THREAD_T	*Thread;
	size_t		i;
	FOM		**fom;
	Cell		**cell;

	if (argc!=9 && argc!=11 && argc!=12 && argc!=13 && argc!=14) Error(
"usage : hp_fdk HiPic/ A B RC Du Ou Dw Ow {{layer1 layer2} RA0 {BPS} TG_format}"
	);
	if ((env=getenv("THREADS"))!=NULL &&
	    (Threads=atoi(env))<=0) Error("bad number of THREADS.");

	InitReadHiPic(argv[1],&hp);

	Nu=hp.Nx; Nw=hp.Ny; M=hp.Nt; for (N1=1; N1<Nu; N1<<=1) ; N2=N1<<1;

	if ((A=atof(argv[2]))<EPS) Error("bad source-sample distance.");
	if ((B=atof(argv[3]))<A) Error("bad source-detector distance.");

	if ((Du=atof(argv[5]))<EPS ||
	    (Dw=atof(argv[7]))<EPS) Error("bad interval of detectors.");

	RC=atof(argv[4]);
	Ou=atof(argv[6]);
	if ((u=RC*B/A)<=(u1=Du*(-Ou)) || u>=(u2=Du*((double)(Nu-1)-Ou)))
	    Error("bad rotation center or origin of detector array.");

	N=(O=(int)floor((r=((r1=(u-u1)/sqrt(B*B+u1*u1))<
			    (r2=(u2-u)/sqrt(B*B+u2*u2)))?r1:r2)*B/Du))*2+1;

	if ((Ow=atof(argv[8]))<=0.0) {
	    l0=(1.0+r)*(-Ow);
	    w=(1.0-r)*(double)(Nw-1)+2.0*r*Ow;
	}
	else {
	    l0=(1.0-r)*(-Ow);
	    w=(Ow<(double)(Nw-1))?(1.0-r)*(double)(Nw-1):
				  (1.0+r)*(double)(Nw-1)-2.0*r*Ow;
	}
	if ((L=(int)ceil(w)+1)<=1) Error("bad origin of detector array.");

	if (argc==13 || argc==14) {
	    if ((z0=atoi(argv[9]))<0 ||
		(z=atoi(argv[10]))>=L || z0>z) Error("bad layer range.");

	    L=z-z0+1; l0+=(double)z0;
	}
	(void)fprintf(stderr,"%d\t%d\t%le\t%le\n",N,L,Dxy=Du*A/B,Dw*A/B);

	if (argc==9) {
	    TermReadHiPic(&hp); return 0;
	}
	if (argc!=11 && argc!=13 && abs(bps=atoi(argv[argc-2]))>sizeof(Cell)*8)
	    Error("bad bits per sample (BPS).");

	if ((Thread=(THREAD_T *)malloc(sizeof(THREAD_T)*Threads))==NULL ||
	    (E=(Vector *)malloc(sizeof(Vector)*N2))==NULL ||
	    (G=(Vector *)malloc(sizeof(Vector)*N2))==NULL ||
	    (V4T=(Vector **)malloc(sizeof(Vector *)*Threads))==NULL ||
	    (*V4T=(Vector *)malloc(sizeof(Vector)*Threads*N2))==NULL ||
	    (F4P=(double **)malloc(sizeof(double *)*Nw))==NULL ||
	    (*F4P=(double *)malloc(sizeof(double)*Nw*Nu))==NULL ||
	    (P=(double **)malloc(sizeof(double *)*Nw))==NULL ||
	    (*P=(double *)malloc(sizeof(double)*Nw*Nu))==NULL ||
	    (Q=(double **)malloc(sizeof(double *)*(Nu+1)))==NULL ||
	    (*Q=(double *)malloc(sizeof(double)*((size_t)Nu*Nw+1)))==NULL ||
	    (S4BP=(Struct **)malloc(sizeof(Struct *)*N))==NULL ||
	    (*S4BP=(Struct *)malloc(sizeof(Struct)*N*N))==NULL ||
	    (F=(Float ***)malloc(sizeof(Float **)*N))==NULL ||
	    (*F=(Float **)malloc(sizeof(Float *)*N*N))==NULL ||
	    (**F=(Float *)malloc(sizeof(Float)*N*N*L))==NULL)
	    Error("memory allocation error.");

	for (j=0; j<N2; j++) {
	    sincos(M_PI*(double)j/(double)N2,&(E[j].y),&(E[j].x));

	    G[j].x=Filter(j-N1); G[j].y=0.0;
	}
	FFT(-1,G);

	for (i=1; i<Threads; i++) V4T[i]=V4T[i-1]+N2;

	for (l=0; l<Nw; l++) {
	    if (l>0) {
		F4P[l]=F4P[l-1]+Nu; P[l]=P[l-1]+Nu;
	    }
	    w=Dw*((double)l-Ow);
	    for (j=0; j<Nu; j++) {
		u=Du*((double)j-Ou);
		F4P[l][j]=(A+RC*u/B)/sqrt(B*B+u*u+w*w)/Du/(double)N2;
	    }
	}
	Q[Nu]=Q[0]; for (j=1; j<Nu; j++) Q[j]=Q[j-1]+Nw;

	O2=O*O;
	for (Y=O, y=0; y<N; y++, Y--) {
	    if (y>0) {
		S4BP[y]=S4BP[y-1]+N; F[y]=F[y-1]+N;
	    }
	    O2_Y2=O2-Y*Y;
	    for (X=(-O), x=0; x<N; x++, X++) {
		S4BP[y][x].j=(O2_Y2>=X*X)-1;

		if (y>0 || x>0) F[y][x]=F[y][x-1]+L;

		for (z=0; z<L; z++) F[y][x][z]=0.0;
	    }
	}
	ra0=atof(argv[(argc==11 || argc==13)?argc-2:argc-3])*M_PI/180.0;

	for (m=0; m<M; m++) {
	    ReadHiPic(&hp,m);

	    for (l=0; l<Nw; l++)
	    for (j=0; j<Nu; j++) P[l][j]=(hp.T[l][j]>0.0)?-log(hp.T[l][j]):0.0;

	    if (m>0)
		for (i=0; i<Threads; i++) TERM_MT(Thread[i],BackProjection);

	    for (i=0; i<Threads; i++) INIT_MT(Thread[i],Convolution,i);

	    sincos(ra0+2.0*M_PI*(double)m/(double)M,&s,&c); s*=Dxy; c*=Dxy;
	    for (Y=  O , y=0; y<N; y++, Y--)
	    for (X=(-O), x=0; x<N; x++, X++)
		if (S4BP[y][x].j>=0) {
		    f=B/(v=A-s*(double)X+c*(double)Y);
		    S4BP[y][x].F4l=A/v;
		    S4BP[y][x].F4Q=f*f*M_PI/(double)M;
		    u=Ou+f*(RC+c*(double)X+s*(double)Y)/Du;
		    S4BP[y][x].u=u-(double)(S4BP[y][x].j=(int)floor(u));
		}

	    for (i=0; i<Threads; i++) TERM_MT(Thread[i],Convolution);
	    for (i=0; i<Threads; i++) INIT_MT(Thread[i],BackProjection,i);
	}
	for (i=0; i<Threads; i++) TERM_MT(Thread[i],BackProjection);

	if (bps==0) {
	    fom=(FOM **)S4BP;

	    for (z=0; z<L; z++) {
		for (y=0; y<N; y++)
		for (x=0; x<N; x++) fom[y][x]=F[y][x][z];

		(void)sprintf(path,argv[argc-1],z+z0);
		StoreImageFile_Float(path,N,N,fom,SIF_F_desc);
		(void)printf("%d\t%s\n",z+z0,SIF_F_desc);
	    }
	}
	else {
	    cell=(Cell **)S4BP;

	    if (bps<0) {
		min=max=F[0][0][0];
		for (y=0; y<N; y++)
		for (x=0; x<N; x++)
		for (z=0; z<L; z++) if (min>F[y][x][z]) min=F[y][x][z]; else
				    if (max<F[y][x][z]) max=F[y][x][z];

		base=min;
		step=(max-min<EPS)?1.0:(max-min)/(double)((1ULL<<(-bps))-1);
		(void)sprintf(SIF_F_desc,SIF_F_FORM,min,max);
		(void)fprintf(stderr,"%s\n",SIF_F_desc);
	    }
	    for (z=0; z<L; z++) {
		min=max=F[0][0][z];
		for (y=0; y<N; y++)
		for (x=0; x<N; x++) if (min>F[y][x][z]) min=F[y][x][z]; else
				    if (max<F[y][x][z]) max=F[y][x][z];

		if (bps>0) {
		    base=min;
		    step=(max-min<EPS)?1.0:(max-min)/(double)((1ULL<<bps)-1);
		    (void)sprintf(SIF_F_desc,SIF_F_FORM,min,max);
		}
		for (y=0; y<N; y++)
		for (x=0; x<N; x++)
		    cell[y][x]=(Cell)((F[y][x][z]-base)/step+0.5);

		(void)sprintf(path,argv[argc-1],z+z0);
		StoreImageFile(path,N,N,abs(bps),cell,SIF_F_desc);
		(void)printf("%d\t" SIF_F_FORM "\n",z+z0,min,max);
	    }
	}
	free(**F); free(*F); free(F); free(*S4BP); free(S4BP);
	free(*Q); free(Q); free(*P); free(P); free(*F4P); free(F4P);
	free(*V4T); free(V4T); free(G); free(E); free(Thread);

	TermReadHiPic(&hp);

	return 0;
}
